/*

Scroller                       ▄ ▄ ▄
for Synchronet                 █████
Version 0.250508               ▐▄█▄▌ cf
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
by Craig Hendricks
codefenix@conchaos.synchro.net

ConstructiveChaos BBS:
  https://conchaos.synchro.net
 telnet://conchaos.synchro.net
    ssh://conchaos.synchro.net

*/

load("sbbsdefs.js");
load("frame.js");
require("graphic.js", 'Graphic');
var sauce = load({},"sauce_lib.js");

const C1 = "\x01w\x01h"; // primary text color
const C2 = "\x01b\x01h"; // secondary color, for symbols and such

const ansi_scrollable_size_limit = 1000000;
const ansi_cursor_movement_limit = 30;

var up_arrow = ascii(24);
var down_arrow = ascii(25);

var contains_ctrl_a = false;

function init_display(opts) { // portions lifted from mrc-client.js
    const w = console.screen_columns;
    const h = console.screen_rows;
    const f = { top: new Frame(1, 1, w, h, BG_BLACK|LIGHTGRAY) };
    f.topbar = new Frame(1, 1, w, 1, BG_BLUE|BLACK, f.top);
    f.viewport = new Frame(1, 2, w, h - 2, BG_BLACK|LIGHTGRAY, f.top);
    f.bottombar = new Frame(1, h, w, 1, BG_BLUE|BLACK, f.top);
    f.viewport.word_wrap = true;
    f.bottombar.putmsg(opts);
    f.top.open();
    return f;
}

function get_lines_for_screen(lines, start, end, x) { 
    if (contains_ctrl_a) {
        var lastCtrlAcodes = "";
        
        // find the last CTRL-A color codes before this line, and
        // prefix it to the return string. This should preserve
        // colors used before the section being shown in the 
        // scroll view.
        
        for (var i = start-1; i >= 0 ; i--) {
            var matched = lines[i].match(/\x01[krgybmcwnh]/ig);
            if (matched) {
                lastCtrlAcodes = matched.join('') + lastCtrlAcodes;                
                if (lastCtrlAcodes.search(/\x01[krgybmcw]/i) >= 0 && lastCtrlAcodes.search(/\x01[nh]/i) >= 0) {                    
                    break;
                }                
            }
        }        
    }
    
    return "\n" + (lastCtrlAcodes ? lastCtrlAcodes : "") + lines.slice(start, end).join("\n") + "\n";
}

function viewfile(file_name, file_title, start_at) {
    var title = file_title ? file_title : file_getname(file_name);
    var exit_viewer = false;
    var line_index;
    var file_lines = [];
    var max_lines;
    var move_viewport = false;
    var ftxt = new File(file_name);
    var sdat = new sauce.read(file_name);
    var ansi_cursor_movement = 0;
    var txt;
    if (ftxt.open("rb")) {       
        txt = ftxt.read().replace(/\f/g, "").replace(/\0/g, " "); // strip out problematic characters
        if (str_is_utf8(txt)) {
            txt = utf8_decode(txt);
        }
        if (txt.search(/\x1a/i) >= 0) {                // only read up to the CTRL Z character, in case 
            txt = txt.substr(0, txt.search(/\x1a/i) ); // of sauce data included in plain text files.
        }
        
        contains_ctrl_a = txt.search(/\x01[krgybmcwnh]/i) >= 0; // Contains CTRL-A codes        
                
        if (txt.indexOf("\x1b[") >= 0) { // Contains ANSI

            // account for sauce data with insane cols/rows values
            var scol = (sdat.cols > 200 ? console.screen_columns : sdat.cols) || console.screen_columns;
            var srow = (sdat.rows > 10 ? 10 : sdat.rows) || 10;
            
            // warn if wider than the terminal
            if (scol > console.screen_columns) {
                console.clear();
                print("\r\n\x01g\x01hAttention!\x01n\r\n\r\nThe ANSI you're about to view is \x01hwider than your terminal's width.");
                print("\x01nCome back with a terminal width of \x01h"+scol+" or more \x01nand try again, or you can view");
                print("it now, but it will be \x01hlimited \x01nto the leftmost \x01h"+console.screen_columns+" \x01ncolumns.\r\n");
                if (console.noyes("View anyway")) {
                    return;
                }
            }
            if (sdat.title) {
                title = '"' + sdat.title.replace(/"/g, '') + '"' + (sdat.author ? (" by " + sdat.author + (sdat.group ? (" of " + sdat.group) : "") ) : "");
            }
            if (sdat.ice_color || file_getext(file_name).toUpperCase()===".ICE" ) {
                print("\x1b[?33;35h"); // switch ice colors on...
                title = title + C1 + " (iCE mode)";
            }        
                       
            if (sdat.tinfos) {
                //log("sdat.tinfos :"+sdat.tinfos);              
                if (sdat.tinfos.search(/amiga|topaz/i) >= 0) { // switch to an "amiga" font if specified.
                    print("\x1b[0;40 D");                      // that's the only font we'll bother with.
                    up_arrow = "up";     // change the up and
                    down_arrow = "down"; // down characters on the bottom bar.
                }
            }    
            
            // Detect animations (i.e.: LOTS of cursor movements), 
            ansi_cursor_movement = (txt.match(/\x1b\[[0-9]*\;[0-9]*H/g) || []).length; 
            
            // Simply dump to console.printfile if the size or cursor movement count exceed the defined limits
            if (txt.length > ansi_scrollable_size_limit || ansi_cursor_movement > ansi_cursor_movement_limit) {
                console.clear(false);
                print("\x1b[;11*r"); // throttle speed down to 115200
                console.printfile(file_name, P_NOPAUSE | P_CPM_EOF | P_NOATCODES);
                print("\x1b[*r"); // reset speed
                console.gotoxy(1, console.screen_rows-1);
                console.pause();
                return;
            }
    
            var graphic = new Graphic(scol, srow);
            graphic.auto_extend = true;
            graphic.ANSI = txt;
            graphic.width = (scol > console.screen_columns ? console.screen_columns : scol)-1;            
            
            file_lines = graphic.MSG.split(/\n/);
        } else {
            file_lines = txt.split(/\n/);
        }        
        ftxt.close();
    }
    const OPTS = " " + // options
     /* up    */ C1 + up_arrow + " " + C2 + "/ " +
     /* down  */ C1 + down_arrow + " " + C2 + "/ " +
     /* pg-up */ C1 + "pg-up " + C2 + "/ " +
     /* pg-dn */ C1 + "pg-dn " + C2 + "/ " +
     /* home  */ C1 + "home " + C2 + "/ " +
     /* end   */ C1 + "end " + C2 + "/ " +
     /* quit  */ C1 + "q" + C2 + ": " + C1 + "quit";
    const frames = init_display(OPTS);
    frames.topbar.putmsg(C2 + "Loading " + C1 + title.substr(0, frames.topbar.width-15) + C2 + "...");
    frames.top.cycle();    
    max_lines = file_lines.length - frames.viewport.height < 1 ? 1 : file_lines.length - frames.viewport.height;
    line_index = start_at === "bottom" ? max_lines : 0;
    frames.viewport.putmsg( get_lines_for_screen(file_lines, line_index, line_index + frames.viewport.height));
    frames.viewport.scrollTo(0, 1);
    frames.bottombar.gotoxy(frames.bottombar.width-6, 1);
    frames.bottombar.putmsg(format( C1+C2+"["+C1+"%.3d%%"+C2+"]", (line_index / max_lines) * 100));
    frames.topbar.clear();
    frames.topbar.putmsg(C1 + "Viewing" + C2 + ": " + C1 + title.substr(0, frames.topbar.width-10));
    frames.top.cycle();
    while (bbs.online && !js.terminated && !exit_viewer) {
        switch (console.inkey().toUpperCase()) {
            case KEY_UP:
                line_index = line_index - 1;
                move_viewport = true;
                break;
            case KEY_DOWN:
                line_index = line_index + 1;
                move_viewport = true;
                break;
            case KEY_PAGEUP:
                line_index = line_index - frames.viewport.height;
                move_viewport = true;
                break;
            case KEY_PAGEDN:
                line_index = line_index + frames.viewport.height;
                move_viewport = true;
                break;
            case KEY_HOME:
                line_index = 0;
                move_viewport = true;
                break;
            case KEY_END:
                line_index = max_lines;
                move_viewport = true;
                break;
            case KEY_ESC:
            case "Q":
                exit_viewer = true;
                break;
        }
        if (move_viewport) {
            if (line_index < 0) {
                line_index = 0;
            } 
            if (line_index > max_lines) {
                line_index = max_lines;
            }
            frames.viewport.clear();
            frames.viewport.putmsg( get_lines_for_screen(file_lines, line_index, line_index + frames.viewport.height) );
            frames.viewport.scrollTo(0, 1);
            frames.viewport.cycle();
            frames.bottombar.gotoxy( frames.bottombar.width-6, 1);
            frames.bottombar.putmsg( format( C1+C2+"["+C1+"%.3d%%"+C2+"]", (line_index / max_lines) * 100) );
            frames.bottombar.cycle();
            move_viewport = false;
        }
        yield();
    }
    frames.top.close();
}

function main(file_name, file_title, scroll_to_top) {
    if (file_exists(file_name)) {
        if (console.type==="ANSI") {
            viewfile(file_name, file_title, scroll_to_top ? scroll_to_top.toLowerCase() : "");
        } else {
            console.printfile(file_name);
        }
    } else {
        log(LOG_ERROR, "File not found: " + file_name);
    }
    
    print("\x1b[?33;35l\x01q\x01l\x01n\x010\x1b[0;0 D"); // Resets iCE colors/blinking, clears screen, resets screen pause, resets colors, and resets font (if changed while rendering an ANSI file).
}

// 1: file path (string)
// 2: title (string; defaults to filename)
// 3: start at (string: top or bottom; defaults to top)

main(argv[0], argv[1], argv[2]);
